import logging
import warnings
from decimal import Decimal
from typing import TYPE_CHECKING, AsyncIterator, Iterable, NamedTuple, Optional, Sequence
from uuid import UUID

import sqlalchemy as sql

from pixeltable import catalog, exprs
from pixeltable.env import Env

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode

if TYPE_CHECKING:
    import pixeltable.plan

_logger = logging.getLogger('pixeltable')


class OrderByItem(NamedTuple):
    expr: exprs.Expr
    asc: Optional[bool]


OrderByClause = list[OrderByItem]


def combine_order_by_clauses(clauses: Iterable[OrderByClause]) -> Optional[OrderByClause]:
    """Returns a clause that's compatible with 'clauses', or None if that doesn't exist.
    Two clauses are compatible if for each of their respective items c1[i] and c2[i]
    a) the exprs are identical and
    b) the asc values are identical or at least one is None (None serves as a wildcard)
    """
    result: OrderByClause = []
    for clause in clauses:
        combined: OrderByClause = []
        for item1, item2 in zip(result, clause):
            if item1.expr.id != item2.expr.id:
                return None
            if item1.asc is not None and item2.asc is not None and item1.asc != item2.asc:
                return None
            asc = item1.asc if item1.asc is not None else item2.asc
            combined.append(OrderByItem(item1.expr, asc))

        # add remaining ordering of the longer list
        prefix_len = min(len(result), len(clause))
        if len(result) > prefix_len:
            combined.extend(result[prefix_len:])
        elif len(clause) > prefix_len:
            combined.extend(clause[prefix_len:])
        result = combined
    return result


def print_order_by_clause(clause: OrderByClause) -> str:
    return ', '.join(
        [
            f'({item.expr}{", asc=True" if item.asc is True else ""}{", asc=False" if item.asc is False else ""})'
            for item in clause
        ]
    )


class SqlNode(ExecNode):
    """
    Materializes data from the store via a Select stmt.
    This only provides the select list. The subclasses are responsible for the From clause and any additional clauses.
    """

    tbl: Optional[catalog.TableVersionPath]
    select_list: exprs.ExprSet
    set_pk: bool
    num_pk_cols: int
    py_filter: Optional[exprs.Expr]  # a predicate that can only be run in Python
    py_filter_eval_ctx: Optional[exprs.RowBuilder.EvalCtx]
    cte: Optional[sql.CTE]
    sql_elements: exprs.SqlElementCache

    # where_clause/-_element: allow subclass to set one or the other (but not both)
    where_clause: Optional[exprs.Expr]
    where_clause_element: Optional[sql.ColumnElement]

    order_by_clause: OrderByClause
    limit: Optional[int]

    def __init__(
        self,
        tbl: Optional[catalog.TableVersionPath],
        row_builder: exprs.RowBuilder,
        select_list: Iterable[exprs.Expr],
        sql_elements: exprs.SqlElementCache,
        set_pk: bool = False,
    ):
        """
        If row_builder contains references to unstored iter columns, expands the select list to include their
        SQL-materializable subexpressions.

        Args:
            select_list: output of the query
            set_pk: if True, sets the primary for each DataRow
        """
        # create Select stmt
        self.sql_elements = sql_elements
        self.tbl = tbl
        assert all(not isinstance(e, exprs.Literal) for e in select_list)  # we're never asked to materialize literals
        self.select_list = exprs.ExprSet(select_list)
        # unstored iter columns: we also need to retrieve whatever is needed to materialize the iter args
        for iter_arg in row_builder.unstored_iter_args.values():
            sql_subexprs = iter_arg.subexprs(filter=self.sql_elements.contains, traverse_matches=False)
            for e in sql_subexprs:
                self.select_list.add(e)
        super().__init__(row_builder, self.select_list, [], None)  # we materialize self.select_list

        if tbl is not None:
            # minimize the number of tables that need to be joined to the target table
            self.retarget_rowid_refs(tbl, self.select_list)

        assert self.sql_elements.contains_all(self.select_list)
        self.set_pk = set_pk
        self.num_pk_cols = 0
        if set_pk:
            # we also need to retrieve the pk columns
            assert tbl is not None
            self.num_pk_cols = len(tbl.tbl_version.get().store_tbl.pk_columns())

        # additional state
        self.result_cursor = None
        # the filter is provided by the subclass
        self.py_filter = None
        self.py_filter_eval_ctx = None
        self.cte = None
        self.limit = None
        self.where_clause = None
        self.where_clause_element = None
        self.order_by_clause = []

    def _create_stmt(self) -> sql.Select:
        """Create Select from local state"""

        assert self.sql_elements.contains_all(self.select_list)
        sql_select_list = [self.sql_elements.get(e) for e in self.select_list]
        if self.set_pk:
            assert self.tbl is not None
            sql_select_list += self.tbl.tbl_version.get().store_tbl.pk_columns()
        stmt = sql.select(*sql_select_list)

        where_clause_element = (
            self.sql_elements.get(self.where_clause) if self.where_clause is not None else self.where_clause_element
        )
        if where_clause_element is not None:
            stmt = stmt.where(where_clause_element)

        order_by_clause: list[sql.ColumnElement] = []
        for e, asc in self.order_by_clause:
            if isinstance(e, exprs.SimilarityExpr):
                order_by_clause.append(e.as_order_by_clause(asc))
            else:
                order_by_clause.append(self.sql_elements.get(e).desc() if asc is False else self.sql_elements.get(e))
        stmt = stmt.order_by(*order_by_clause)

        if self.py_filter is None and self.limit is not None:
            # if we don't have a Python filter, we can apply the limit to stmt
            stmt = stmt.limit(self.limit)

        return stmt

    def _ordering_tbl_ids(self) -> set[UUID]:
        return exprs.Expr.all_tbl_ids(e for e, _ in self.order_by_clause)

    def to_cte(self) -> Optional[tuple[sql.CTE, exprs.ExprDict[sql.ColumnElement]]]:
        """
        Returns a CTE that materializes the output of this node plus a mapping from select list expr to output column

        Returns:
            (CTE, dict from Expr to output column)
        """
        if self.py_filter is not None:
            # the filter needs to run in Python
            return None
        self.set_pk = False  # we don't need the PK if we use this SqlNode as a CTE
        if self.cte is None:
            self.cte = self._create_stmt().cte()
            assert len(self.cte.c) == len(self.select_list)
        return self.cte, exprs.ExprDict(zip(self.select_list, self.cte.c))

    @classmethod
    def retarget_rowid_refs(cls, target: catalog.TableVersionPath, expr_seq: Iterable[exprs.Expr]) -> None:
        """Change rowid refs to point to target"""
        for e in expr_seq:
            if isinstance(e, exprs.RowidRef):
                e.set_tbl(target)

    @classmethod
    def create_from_clause(
        cls,
        tbl: catalog.TableVersionPath,
        stmt: sql.Select,
        refd_tbl_ids: Optional[set[UUID]] = None,
        exact_version_only: Optional[set[UUID]] = None,
    ) -> sql.Select:
        """Add From clause to stmt for tables/views referenced by materialized_exprs
        Args:
            tbl: root table of join chain
            stmt: stmt to add From clause to
            materialized_exprs: list of exprs that reference tables in the join chain; if empty, include only the root
            exact_version_only: set of table ids for which we only want to see rows created at the current version
        Returns:
            augmented stmt
        """
        # we need to include at least the root
        if refd_tbl_ids is None:
            refd_tbl_ids = set()
        if exact_version_only is None:
            exact_version_only = set()
        candidates = tbl.get_tbl_versions()
        assert len(candidates) > 0
        joined_tbls: list[catalog.TableVersionHandle] = [candidates[0]]
        for t in candidates[1:]:
            if t.id in refd_tbl_ids:
                joined_tbls.append(t)

        first = True
        prev_tbl: Optional[catalog.TableVersionHandle] = None
        for t in joined_tbls[::-1]:
            if first:
                stmt = stmt.select_from(t.get().store_tbl.sa_tbl)
                first = False
            else:
                # join tbl to prev_tbl on prev_tbl's rowid cols
                prev_tbl_rowid_cols = prev_tbl.get().store_tbl.rowid_columns()
                tbl_rowid_cols = t.get().store_tbl.rowid_columns()
                rowid_clauses = [
                    c1 == c2 for c1, c2 in zip(prev_tbl_rowid_cols, tbl_rowid_cols[: len(prev_tbl_rowid_cols)])
                ]
                stmt = stmt.join(t.get().store_tbl.sa_tbl, sql.and_(*rowid_clauses))
            if t.id in exact_version_only:
                stmt = stmt.where(t.get().store_tbl.v_min_col == t.get().version)
            else:
                stmt = stmt.where(t.get().store_tbl.v_min_col <= t.get().version).where(
                    t.get().store_tbl.v_max_col > t.get().version
                )
            prev_tbl = t
        return stmt

    def set_where(self, where_clause: exprs.Expr) -> None:
        assert self.where_clause_element is None
        self.where_clause = where_clause

    def set_py_filter(self, py_filter: exprs.Expr) -> None:
        assert self.py_filter is None
        self.py_filter = py_filter
        self.py_filter_eval_ctx = self.row_builder.create_eval_ctx([py_filter], exclude=self.select_list)

    def set_order_by(self, ordering: OrderByClause) -> None:
        """Add Order By clause"""
        if self.tbl is not None:
            # change rowid refs against a base table to rowid refs against the target table, so that we minimize
            # the number of tables that need to be joined to the target table
            self.retarget_rowid_refs(self.tbl, [e for e, _ in ordering])
        combined = combine_order_by_clauses([self.order_by_clause, ordering])
        assert combined is not None
        self.order_by_clause = combined

    def set_limit(self, limit: int) -> None:
        self.limit = limit

    def _log_explain(self, stmt: sql.Select) -> None:
        conn = Env.get().conn
        try:
            # don't set dialect=Env.get().engine.dialect: x % y turns into x %% y, which results in a syntax error
            stmt_str = str(stmt.compile(compile_kwargs={'literal_binds': True}))
            explain_result = conn.execute(sql.text(f'EXPLAIN {stmt_str}'))
            explain_str = '\n'.join([str(row) for row in explain_result])
            _logger.debug(f'SqlScanNode explain:\n{explain_str}')
        except Exception as e:
            _logger.warning(f'EXPLAIN failed with error: {e}')

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        # run the query; do this here rather than in _open(), exceptions are only expected during iteration
        with warnings.catch_warnings(record=True) as w:
            stmt = self._create_stmt()
            try:
                # log stmt, if possible
                stmt_str = str(stmt.compile(compile_kwargs={'literal_binds': True}))
                _logger.debug(f'SqlLookupNode stmt:\n{stmt_str}')
            except Exception:
                pass
            self._log_explain(stmt)

            conn = Env.get().conn
            result_cursor = conn.execute(stmt)
            for _ in w:
                pass

        tbl_version = self.tbl.tbl_version if self.tbl is not None else None
        output_batch = DataRowBatch(tbl_version, self.row_builder)
        output_row: Optional[exprs.DataRow] = None
        num_rows_returned = 0

        for sql_row in result_cursor:
            output_row = output_batch.add_row(output_row)

            # populate output_row
            if self.num_pk_cols > 0:
                output_row.set_pk(tuple(sql_row[-self.num_pk_cols :]))
            # copy the output of the SQL query into the output row
            for i, e in enumerate(self.select_list):
                slot_idx = e.slot_idx
                # certain numerical operations can produce Decimals (eg, SUM(<int column>)); we need to convert them
                if isinstance(sql_row[i], Decimal):
                    if e.col_type.is_int_type():
                        output_row[slot_idx] = int(sql_row[i])
                    elif e.col_type.is_float_type():
                        output_row[slot_idx] = float(sql_row[i])
                    else:
                        raise RuntimeError(f'Unexpected Decimal value for {e}')
                else:
                    output_row[slot_idx] = sql_row[i]

            if self.py_filter is not None:
                # evaluate filter
                self.row_builder.eval(output_row, self.py_filter_eval_ctx, profile=self.ctx.profile)
            if self.py_filter is not None and not output_row[self.py_filter.slot_idx]:
                # we re-use this row for the next sql row since it didn't pass the filter
                output_row = output_batch.pop_row()
                output_row.clear()
            else:
                # reset output_row in order to add new one
                output_row = None
                num_rows_returned += 1

            if self.limit is not None and num_rows_returned == self.limit:
                break

            if self.ctx.batch_size > 0 and len(output_batch) == self.ctx.batch_size:
                _logger.debug(f'SqlScanNode: returning {len(output_batch)} rows')
                yield output_batch
                output_batch = DataRowBatch(tbl_version, self.row_builder)

        if len(output_batch) > 0:
            _logger.debug(f'SqlScanNode: returning {len(output_batch)} rows')
            yield output_batch

    def _close(self) -> None:
        if self.result_cursor is not None:
            self.result_cursor.close()


class SqlScanNode(SqlNode):
    """
    Materializes data from the store via a Select stmt.

    Supports filtering and ordering.
    """

    exact_version_only: list[catalog.TableVersionHandle]

    def __init__(
        self,
        tbl: catalog.TableVersionPath,
        row_builder: exprs.RowBuilder,
        select_list: Iterable[exprs.Expr],
        set_pk: bool = False,
        exact_version_only: Optional[list[catalog.TableVersionHandle]] = None,
    ):
        """
        Args:
            select_list: output of the query
            set_pk: if True, sets the primary for each DataRow
            exact_version_only: tables for which we only want to see rows created at the current version
        """
        sql_elements = exprs.SqlElementCache()
        super().__init__(tbl, row_builder, select_list, sql_elements, set_pk=set_pk)
        # create Select stmt
        if exact_version_only is None:
            exact_version_only = []

        self.exact_version_only = exact_version_only

    def _create_stmt(self) -> sql.Select:
        stmt = super()._create_stmt()
        where_clause_tbl_ids = self.where_clause.tbl_ids() if self.where_clause is not None else set()
        refd_tbl_ids = exprs.Expr.all_tbl_ids(self.select_list) | where_clause_tbl_ids | self._ordering_tbl_ids()
        stmt = self.create_from_clause(
            self.tbl, stmt, refd_tbl_ids, exact_version_only={t.id for t in self.exact_version_only}
        )
        return stmt


class SqlLookupNode(SqlNode):
    """
    Materializes data from the store via a Select stmt with a WHERE clause that matches a list of key values
    """

    def __init__(
        self,
        tbl: catalog.TableVersionPath,
        row_builder: exprs.RowBuilder,
        select_list: Iterable[exprs.Expr],
        sa_key_cols: list[sql.Column],
        key_vals: list[tuple],
    ):
        """
        Args:
            select_list: output of the query
            sa_key_cols: list of key columns in the store table
            key_vals: list of key values to look up
        """
        sql_elements = exprs.SqlElementCache()
        super().__init__(tbl, row_builder, select_list, sql_elements, set_pk=True)
        # Where clause: (key-col-1, key-col-2, ...) IN ((val-1, val-2, ...), ...)
        self.where_clause_element = sql.tuple_(*sa_key_cols).in_(key_vals)

    def _create_stmt(self) -> sql.Select:
        stmt = super()._create_stmt()
        refd_tbl_ids = exprs.Expr.all_tbl_ids(self.select_list) | self._ordering_tbl_ids()
        stmt = self.create_from_clause(self.tbl, stmt, refd_tbl_ids)
        return stmt


class SqlAggregationNode(SqlNode):
    """
    Materializes data from the store via a Select stmt with a WHERE clause that matches a list of key values
    """

    group_by_items: Optional[list[exprs.Expr]]

    def __init__(
        self,
        row_builder: exprs.RowBuilder,
        input: SqlNode,
        select_list: Iterable[exprs.Expr],
        group_by_items: Optional[list[exprs.Expr]] = None,
        limit: Optional[int] = None,
        exact_version_only: Optional[list[catalog.TableVersion]] = None,
    ):
        """
        Args:
            select_list: can contain calls to AggregateFunctions
            group_by_items: list of expressions to group by
            limit: max number of rows to return: None = no limit
        """
        _, input_col_map = input.to_cte()
        sql_elements = exprs.SqlElementCache(input_col_map)
        super().__init__(None, row_builder, select_list, sql_elements)
        self.group_by_items = group_by_items

    def _create_stmt(self) -> sql.Select:
        stmt = super()._create_stmt()
        if self.group_by_items is not None:
            sql_group_by_items = [self.sql_elements.get(e) for e in self.group_by_items]
            assert all(e is not None for e in sql_group_by_items)
            stmt = stmt.group_by(*sql_group_by_items)
        return stmt


class SqlJoinNode(SqlNode):
    """
    Materializes data from the store via a Select ... From ... that contains joins
    """

    input_ctes: list[sql.CTE]
    join_clauses: list['pixeltable.plan.JoinClause']

    def __init__(
        self,
        row_builder: exprs.RowBuilder,
        inputs: Sequence[SqlNode],
        join_clauses: list['pixeltable.plan.JoinClause'],
        select_list: Iterable[exprs.Expr],
    ):
        assert len(inputs) > 1
        assert len(inputs) == len(join_clauses) + 1
        self.input_ctes = []
        self.join_clauses = join_clauses
        sql_elements = exprs.SqlElementCache()
        for input_node in inputs:
            input_cte, input_col_map = input_node.to_cte()
            self.input_ctes.append(input_cte)
            sql_elements.extend(input_col_map)
        super().__init__(None, row_builder, select_list, sql_elements)

    def _create_stmt(self) -> sql.Select:
        from pixeltable import plan

        stmt = super()._create_stmt()
        stmt = stmt.select_from(self.input_ctes[0])
        for i in range(len(self.join_clauses)):
            join_clause = self.join_clauses[i]
            on_clause = (
                self.sql_elements.get(join_clause.join_predicate)
                if join_clause.join_type != plan.JoinType.CROSS
                else sql.sql.expression.literal(True)
            )
            is_outer = join_clause.join_type in (plan.JoinType.LEFT, plan.JoinType.FULL_OUTER)
            stmt = stmt.join(
                self.input_ctes[i + 1],
                onclause=on_clause,
                isouter=is_outer,
                full=join_clause == plan.JoinType.FULL_OUTER,
            )
        return stmt



from __future__ import annotations

import asyncio
import logging
import traceback
from types import TracebackType
from typing import AsyncIterator, Iterable, Optional, Union

import numpy as np

import pixeltable.exceptions as excs
from pixeltable import exprs

from ..data_row_batch import DataRowBatch
from ..exec_node import ExecNode
from .evaluators import FnCallEvaluator, NestedRowList
from .globals import ExecCtx, Scheduler
from .row_buffer import RowBuffer
from .schedulers import SCHEDULERS

_logger = logging.getLogger('pixeltable')


class ExprEvalNode(ExecNode):
    """
    Expression evaluation

    Resource management:
    - the execution system tries to limit total memory consumption by limiting the number of rows that are in
      circulation
    - during execution, slots that aren't part of the output are garbage collected as soon as their direct dependents
      are materialized

    TODO:
    - Literal handling: currently, Literal values are copied into slots via the normal evaluation mechanism, which is
      needless overhead; instead: pre-populate Literal slots in _init_row()
    - dynamically determine MAX_BUFFERED_ROWS, based on the avg memory consumption of a row and our configured memory
      limit
    - local model inference on gpu: currently, no attempt is made to ensure that models can fit onto the gpu
      simultaneously, which will cause errors; instead, the execution should be divided into sequential phases, each
      of which only contains a subset of the models which is known to fit onto the gpu simultaneously
    """

    maintain_input_order: bool  # True if we're returning rows in the order we received them from our input
    outputs: np.ndarray  # bool per slot; True if this slot is part of our output
    schedulers: dict[str, Scheduler]  # key: resource pool name
    exec_ctx: ExecCtx  # for input/output rows

    # execution state
    tasks: set[asyncio.Task]  # collects all running tasks to prevent them from getting gc'd
    exc_event: asyncio.Event  # set if an exception needs to be propagated
    error: Optional[Union[excs.Error, excs.ExprEvalError]]  # exception that needs to be propagated
    completed_rows: asyncio.Queue[exprs.DataRow]  # rows that have completed evaluation
    completed_event: asyncio.Event  # set when completed_rows is non-empty
    input_iter: AsyncIterator[DataRowBatch]
    current_input_batch: Optional[DataRowBatch]  # batch from which we're currently consuming rows
    input_row_idx: int  # next row to consume from current_input_batch
    next_input_batch: Optional[DataRowBatch]  # read-ahead input batch
    avail_input_rows: int  # total number across both current_/next_input_batch
    input_complete: bool  # True if we've received all input batches
    num_in_flight: int  # number of dispatched rows that haven't completed
    row_pos_map: Optional[dict[int, int]]  # id(row) -> position of row in input; only set if maintain_input_order
    output_buffer: RowBuffer  # holds rows that are ready to be returned, in order

    # debugging
    num_input_rows: int
    num_output_rows: int

    BATCH_SIZE = 64
    MAX_BUFFERED_ROWS = 2048  # maximum number of rows that have been dispatched but not yet returned

    def __init__(
        self,
        row_builder: exprs.RowBuilder,
        output_exprs: Iterable[exprs.Expr],
        input_exprs: Iterable[exprs.Expr],
        input: ExecNode,
        maintain_input_order: bool = True,
    ):
        super().__init__(row_builder, output_exprs, input_exprs, input)
        self.maintain_input_order = maintain_input_order
        self.outputs = np.zeros(row_builder.num_materialized, dtype=bool)
        output_slot_idxs = [e.slot_idx for e in output_exprs]
        self.outputs[output_slot_idxs] = True
        self.tasks = set()
        self.error = None

        self.input_iter = self.input.__aiter__()
        self.current_input_batch = None
        self.next_input_batch = None
        self.input_row_idx = 0
        self.avail_input_rows = 0
        self.input_complete = False
        self.num_in_flight = 0
        self.row_pos_map = None
        self.output_buffer = RowBuffer(self.MAX_BUFFERED_ROWS)

        self.num_input_rows = 0
        self.num_output_rows = 0

        # self.slot_evaluators = {}
        self.schedulers = {}
        # self._init_slot_evaluators()
        self.exec_ctx = ExecCtx(self, self.row_builder, output_exprs, input_exprs)

    def set_input_order(self, maintain_input_order: bool) -> None:
        self.maintain_input_order = maintain_input_order

    async def _fetch_input_batch(self) -> None:
        """
        Fetches another batch from our input or sets input_complete to True if there are no more batches.

        - stores the batch in current_input_batch, if not already set, or next_input_batch
        - updates row_pos_map, if needed
        """
        assert not self.input_complete
        try:
            batch = await self.input_iter.__anext__()
            assert self.next_input_batch is None
            if self.current_input_batch is None:
                self.current_input_batch = batch
            else:
                self.next_input_batch = batch
            if self.maintain_input_order:
                for idx, row in enumerate(batch.rows):
                    self.row_pos_map[id(row)] = self.num_input_rows + idx
            self.num_input_rows += len(batch)
            self.avail_input_rows += len(batch)
            _logger.debug(
                f'adding input: batch_size={len(batch)} #input_rows={self.num_input_rows} '
                f'#avail={self.avail_input_rows}'
            )
        except StopAsyncIteration:
            self.input_complete = True
            _logger.debug(f'finished input: #input_rows={self.num_input_rows}, #avail={self.avail_input_rows}')
        except excs.Error as err:
            self.error = err
            self.exc_event.set()
        # TODO: should we also handle Exception here and create an excs.Error from it?

    @property
    def total_buffered(self) -> int:
        return self.num_in_flight + self.completed_rows.qsize() + self.output_buffer.num_rows

    def _dispatch_input_rows(self) -> None:
        """Dispatch the maximum number of input rows, given total_buffered; does not block"""
        if self.avail_input_rows == 0:
            return
        num_rows = min(self.MAX_BUFFERED_ROWS - self.total_buffered, self.avail_input_rows)
        assert num_rows >= 0
        if num_rows == 0:
            return
        assert self.current_input_batch is not None
        avail_current_batch_rows = len(self.current_input_batch) - self.input_row_idx

        rows: list[exprs.DataRow]
        if avail_current_batch_rows > num_rows:
            # we only need rows from current_input_batch
            rows = self.current_input_batch.rows[self.input_row_idx : self.input_row_idx + num_rows]
            self.input_row_idx += num_rows
        else:
            # we need rows from both current_/next_input_batch
            rows = self.current_input_batch.rows[self.input_row_idx :]
            self.current_input_batch = self.next_input_batch
            self.next_input_batch = None
            self.input_row_idx = 0
            num_remaining = num_rows - len(rows)
            if num_remaining > 0:
                rows.extend(self.current_input_batch.rows[:num_remaining])
                self.input_row_idx = num_remaining
        self.avail_input_rows -= num_rows
        self.num_in_flight += num_rows
        self._log_state(f'dispatch input ({num_rows})')

        self.exec_ctx.init_rows(rows)
        self.dispatch(rows, self.exec_ctx)

    def _log_state(self, prefix: str) -> None:
        _logger.debug(
            f'{prefix}: #in-flight={self.num_in_flight} #complete={self.completed_rows.qsize()} '
            f'#output-buffer={self.output_buffer.num_rows} #ready={self.output_buffer.num_ready} '
            f'total-buffered={self.total_buffered} #avail={self.avail_input_rows} '
            f'#input={self.num_input_rows} #output={self.num_output_rows}'
        )

    def _init_schedulers(self) -> None:
        resource_pools = {
            eval.fn_call.resource_pool
            for eval in self.exec_ctx.slot_evaluators.values()
            if isinstance(eval, FnCallEvaluator)
        }
        resource_pools = {pool for pool in resource_pools if pool is not None}
        for pool_name in resource_pools:
            for scheduler in SCHEDULERS:
                if scheduler.matches(pool_name):
                    self.schedulers[pool_name] = scheduler(pool_name, self)
                    break
            if pool_name not in self.schedulers:
                raise RuntimeError(f'No scheduler found for resource pool {pool_name}')

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        """
        Main event loop

        Goals:
        - return completed DataRowBatches as soon as they become available
        - maximize the number of rows in flight in order to maximize parallelism, up to the given limit
        """
        # initialize completed_rows and events, now that we have the correct event loop
        self.completed_rows = asyncio.Queue[exprs.DataRow]()
        self.exc_event = asyncio.Event()
        self.completed_event = asyncio.Event()
        self._init_schedulers()
        if self.maintain_input_order:
            self.row_pos_map = {}
            self.output_buffer.set_row_pos_map(self.row_pos_map)

        row: exprs.DataRow
        exc_event_aw = asyncio.create_task(self.exc_event.wait(), name='exc_event.wait()')
        input_batch_aw: Optional[asyncio.Task] = None
        completed_aw: Optional[asyncio.Task] = None
        closed_evaluators = False  # True after calling Evaluator.close()

        try:
            while True:
                # process completed rows before doing anything else
                while not self.completed_rows.empty():
                    # move completed rows to output buffer
                    while not self.completed_rows.empty():
                        row = self.completed_rows.get_nowait()
                        self.output_buffer.add_row(row)
                        if self.row_pos_map is not None:
                            self.row_pos_map.pop(id(row))

                    self._log_state('processed completed')
                    # return as many batches as we have available
                    while self.output_buffer.num_ready >= self.BATCH_SIZE:
                        batch_rows = self.output_buffer.get_rows(self.BATCH_SIZE)
                        self.num_output_rows += len(batch_rows)
                        # make sure we top up our in-flight rows before yielding
                        self._dispatch_input_rows()
                        self._log_state(f'yielding {len(batch_rows)} rows')
                        yield DataRowBatch(tbl=None, row_builder=self.row_builder, rows=batch_rows)
                        # at this point, we may have more completed rows

                assert self.completed_rows.empty()  # all completed rows should be sitting in output_buffer
                self.completed_event.clear()
                if self.input_complete and self.num_in_flight == 0:
                    # there is no more input and nothing left to wait for
                    assert self.avail_input_rows == 0
                    if self.output_buffer.num_ready > 0:
                        assert self.output_buffer.num_rows == self.output_buffer.num_ready
                        # yield the leftover rows
                        batch_rows = self.output_buffer.get_rows(self.output_buffer.num_ready)
                        self.num_output_rows += len(batch_rows)
                        self._log_state(f'yielding {len(batch_rows)} rows')
                        yield DataRowBatch(tbl=None, row_builder=self.row_builder, rows=batch_rows)

                    assert self.output_buffer.num_rows == 0
                    return

                if self.input_complete and self.avail_input_rows == 0 and not closed_evaluators:
                    # no more input rows to dispatch, but we're still waiting for rows to finish:
                    # close  all slot evaluators to flush queued rows
                    for evaluator in self.exec_ctx.slot_evaluators.values():
                        evaluator.close()
                    closed_evaluators = True

                # we don't have a full batch of rows at this point and need to wait
                aws = {exc_event_aw}  # always wait for an exception
                if self.next_input_batch is None and not self.input_complete:
                    # also wait for another batch if we don't have a read-ahead batch yet
                    if input_batch_aw is None:
                        input_batch_aw = asyncio.create_task(self._fetch_input_batch(), name='_fetch_input_batch()')
                    aws.add(input_batch_aw)
                if self.num_in_flight > 0:
                    # also wait for more rows to complete
                    if completed_aw is None:
                        completed_aw = asyncio.create_task(self.completed_event.wait(), name='completed.wait()')
                    aws.add(completed_aw)
                done, _ = await asyncio.wait(aws, return_when=asyncio.FIRST_COMPLETED)

                if self.exc_event.is_set():
                    # we got an exception that we need to propagate through __iter__()
                    _logger.debug(f'Propagating exception {self.error}')
                    if isinstance(self.error, excs.ExprEvalError):
                        raise self.error from self.error.exc
                    else:
                        raise self.error
                if completed_aw in done:
                    self._log_state('completed_aw done')
                    completed_aw = None
                if input_batch_aw in done:
                    self._dispatch_input_rows()
                    input_batch_aw = None

        finally:
            # task cleanup
            active_tasks = {exc_event_aw}
            if input_batch_aw is not None:
                active_tasks.add(input_batch_aw)
            if completed_aw is not None:
                active_tasks.add(completed_aw)
            active_tasks.update(self.tasks)
            for task in active_tasks:
                if not task.done():
                    task.cancel()
            _ = await asyncio.gather(*active_tasks, return_exceptions=True)

    def dispatch_exc(
        self, rows: list[exprs.DataRow], slot_with_exc: int, exc_tb: TracebackType, exec_ctx: ExecCtx
    ) -> None:
        """Propagate exception to main event loop or to dependent slots, depending on ignore_errors"""
        if len(rows) == 0 or self.exc_event.is_set():
            return

        if not self.ctx.ignore_errors:
            dependency_idxs = [e.slot_idx for e in exec_ctx.row_builder.unique_exprs[slot_with_exc].dependencies()]
            first_row = rows[0]
            input_vals = [first_row[idx] for idx in dependency_idxs]
            e = exec_ctx.row_builder.unique_exprs[slot_with_exc]
            self.error = excs.ExprEvalError(e, f'expression {e}', first_row.get_exc(e.slot_idx), exc_tb, input_vals, 0)
            self.exc_event.set()
            return

        for row in rows:
            assert row.has_exc(slot_with_exc)
            exc = row.get_exc(slot_with_exc)
            # propagate exception
            for slot_idx in np.nonzero(exec_ctx.row_builder.transitive_dependents[slot_with_exc])[0].tolist():
                row.set_exc(slot_idx, exc)
        self.dispatch(rows, exec_ctx)

    def dispatch(self, rows: list[exprs.DataRow], exec_ctx: ExecCtx) -> None:
        """Dispatch rows to slot evaluators, based on materialized dependencies"""
        if len(rows) == 0 or self.exc_event.is_set():
            return

        # slots ready for evaluation; rows x slots
        ready_slots = np.zeros((len(rows), exec_ctx.row_builder.num_materialized), dtype=bool)
        completed_rows = np.zeros(len(rows), dtype=bool)
        for i, row in enumerate(rows):
            row.missing_slots &= row.has_val == False
            if row.missing_slots.sum() == 0:
                # all output slots have been materialized
                completed_rows[i] = True
            else:
                # dependencies of missing slots
                missing_dependencies = exec_ctx.row_builder.num_dependencies * row.missing_slots
                # determine ready slots that are not yet materialized and not yet scheduled
                num_mat_dependencies = np.sum(exec_ctx.row_builder.dependencies * row.has_val, axis=1)
                num_missing = missing_dependencies - num_mat_dependencies
                ready_slots[i] = (num_missing == 0) & (row.is_scheduled == False) & row.missing_slots
                row.is_scheduled |= ready_slots[i]

            # clear intermediate values that are no longer needed (ie, all dependents are materialized)
            missing_dependents = np.sum(exec_ctx.row_builder.dependencies[row.has_val == False], axis=0)
            gc_targets = (missing_dependents == 0) & (row.missing_dependents > 0) & exec_ctx.gc_targets
            row.clear(gc_targets)
            row.missing_dependents = missing_dependents

        if np.any(completed_rows):
            completed_idxs = list(completed_rows.nonzero()[0])
            if rows[i].parent_row is not None:
                # these are nested rows
                for i in completed_idxs:
                    row = rows[i]
                    assert row.parent_row is not None and row.parent_slot_idx is not None
                    assert isinstance(row.parent_row.vals[row.parent_slot_idx], NestedRowList)
                    row.parent_row.vals[row.parent_slot_idx].complete_row()
            else:
                for i in completed_idxs:
                    self.completed_rows.put_nowait(rows[i])
                self.completed_event.set()
                self.num_in_flight -= len(completed_idxs)

        # schedule all ready slots
        for slot_idx in np.sum(ready_slots, axis=0).nonzero()[0]:
            ready_rows_v = ready_slots[:, slot_idx].flatten()
            _ = ready_rows_v.nonzero()
            ready_rows = [rows[i] for i in ready_rows_v.nonzero()[0]]
            _logger.debug(f'Scheduling {len(ready_rows)} rows for slot {slot_idx}')
            exec_ctx.slot_evaluators[slot_idx].schedule(ready_rows, slot_idx)

    def register_task(self, t: asyncio.Task) -> None:
        self.tasks.add(t)
        t.add_done_callback(self._done_cb)

    def _done_cb(self, t: asyncio.Task) -> None:
        self.tasks.discard(t)
        # end the main loop if we had an unhandled exception
        try:
            t.result()
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            stack_trace = traceback.format_exc()
            self.error = excs.Error(f'Exception in task: {exc}\n{stack_trace}')
            self.exc_event.set()



from typing import Any, Optional

import sqlalchemy as sql

import pixeltable.type_system as ts
from pixeltable.exprs import Expr, FunctionCall
from pixeltable.func import FunctionRegistry

from .data_row import DataRow
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class MethodRef(Expr):
    """
    A method reference. This represents a `Function` instance with its
    first parameter bound to a base expression.

    When a `MethodRef` is called, it returns a `FunctionCall` with the base expression as the first argument.
    The effective arity of a `MethodRef` is one less than the arity of the underlying `Function`.
    """

    # TODO: Should this even be an `Expr`? It can't actually be evaluated directly (it has to be first
    #   converted to a `FunctionCall` by binding any remaining parameters).

    def __init__(self, base_expr: Expr, method_name: str) -> None:
        super().__init__(ts.InvalidType())  # The `MethodRef` is untyped until it is called.
        self.base_expr = base_expr
        self.method_name = method_name
        self.fn = FunctionRegistry.get().lookup_type_method(base_expr.col_type.type_enum, method_name)
        if self.fn is None:
            # This has to be an `AttributeError`, or tab-completion won't work properly in ipython.
            raise AttributeError(f'Unknown method (of type {base_expr.col_type}): {method_name}')
        self.components = [base_expr]
        self.id = self._create_id()

    def _as_dict(self) -> dict:
        return {'method_name': self.method_name, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> 'MethodRef':
        assert 'method_name' in d
        assert len(components) == 1
        return cls(components[0], d['method_name'])

    def __call__(self, *args: Any, **kwargs: Any) -> FunctionCall:
        result = self.fn(*[self.base_expr, *args], **kwargs)
        assert isinstance(result, FunctionCall)
        result.is_method_call = True
        return result

    def _equals(self, other: 'MethodRef') -> bool:
        return self.base_expr.id == other.base_expr.id and self.method_name == other.method_name

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('method_name', self.method_name)]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        raise AssertionError('MethodRef cannot be evaluated directly')

    def __repr__(self) -> str:
        return f'{self.base_expr}.{self.method_name}'



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Fireworks AI API. In order to use them, you must
first `pip install fireworks-ai` and configure your Fireworks AI credentials, as described in
the [Working with Fireworks](https://pixeltable.readme.io/docs/working-with-fireworks) tutorial.
"""

from typing import TYPE_CHECKING, Optional

import pixeltable as pxt
from pixeltable import env
from pixeltable.config import Config
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import fireworks.client  # type: ignore[import-untyped]


@env.register_client('fireworks')
def _(api_key: str) -> 'fireworks.client.Fireworks':
    import fireworks.client

    return fireworks.client.Fireworks(api_key=api_key)


def _fireworks_client() -> 'fireworks.client.Fireworks':
    return env.Env.get().get_client('fireworks')


@pxt.udf(resource_pool='request-rate:fireworks')
async def chat_completions(
    messages: list[dict[str, str]],
    *,
    model: str,
    max_tokens: Optional[int] = None,
    top_k: Optional[int] = None,
    top_p: Optional[float] = None,
    temperature: Optional[float] = None,
    request_timeout: Optional[int] = None,
) -> dict:
    """
    Creates a model response for the given chat conversation.

    Equivalent to the Fireworks AI `chat/completions` API endpoint.
    For additional details, see: <https://docs.fireworks.ai/api-reference/post-chatcompletions>

    Request throttling:
    Applies the rate limit set in the config (section `fireworks`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install fireworks-ai`

    Args:
        messages: A list of messages comprising the conversation so far.
        model: The name of the model to use.

    For details on the other parameters, see: <https://docs.fireworks.ai/api-reference/post-chatcompletions>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `accounts/fireworks/models/mixtral-8x22b-instruct`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> messages = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(
        ...     response=chat_completions(messages, model='accounts/fireworks/models/mixtral-8x22b-instruct')
        ... )
    """
    kwargs = {'max_tokens': max_tokens, 'top_k': top_k, 'top_p': top_p, 'temperature': temperature}
    kwargs_not_none = {k: v for k, v in kwargs.items() if v is not None}

    # for debugging purposes:
    # res_sync = _fireworks_client().chat.completions.create(model=model, messages=messages, **kwargs_not_none)
    # res_sync_dict = res_sync.dict()

    if request_timeout is None:
        request_timeout = Config.get().get_int_value('timeout', section='fireworks') or 600
    # TODO: this timeout doesn't really work, I think it only applies to returning the stream, but not to the timing
    # of the chunks; addressing this would require a timeout for the task running this udf
    stream = _fireworks_client().chat.completions.acreate(
        model=model, messages=messages, request_timeout=request_timeout, **kwargs_not_none
    )
    chunks = []
    async for chunk in stream:
        chunks.append(chunk)

    res = {
        'id': chunks[0].id,
        'object': 'chat.completion',
        'created': chunks[0].created,
        'model': chunks[0].model,
        'choices': [
            {
                'index': 0,
                'message': {
                    'role': None,
                    'content': '',
                    'tool_calls': None,
                    'tool_call_id': None,
                    'function': None,
                    'name': None,
                },
                'finish_reason': None,
                'logprobs': None,
                'raw_output': None,
            }
        ],
        'usage': {},
    }
    for chunk in chunks:
        d = chunk.dict()
        if 'usage' in d and d['usage'] is not None:
            res['usage'] = d['usage']
        if chunk.choices[0].finish_reason is not None:
            res['choices'][0]['finish_reason'] = chunk.choices[0].finish_reason
        if chunk.choices[0].delta.role is not None:
            res['choices'][0]['message']['role'] = chunk.choices[0].delta.role
        if chunk.choices[0].delta.content is not None:
            res['choices'][0]['message']['content'] += chunk.choices[0].delta.content
        if chunk.choices[0].delta.tool_calls is not None:
            res['choices'][0]['message']['tool_calls'] = chunk.choices[0].delta.tool_calls
        if chunk.choices[0].delta.function is not None:
            res['choices'][0]['message']['function'] = chunk.choices[0].delta.function
    return res


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from uuid import UUID

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=17)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, table_md_updater=__update_table_md)


def __update_table_md(table_md: dict, table_id: UUID) -> None:
    # key changes in IndexMd.init_args: img_embed -> image_embed, txt_embed -> string_embed
    if len(table_md['index_md']) == 0:
        return
    for idx_md in table_md['index_md'].values():
        if not idx_md['class_fqn'].endswith('.EmbeddingIndex'):
            continue
        init_dict = idx_md['init_args']
        init_dict['image_embed'] = init_dict['img_embed']
        del init_dict['img_embed']
        init_dict['string_embed'] = init_dict['txt_embed']
        del init_dict['txt_embed']



from __future__ import annotations

import enum
from typing import Any, ClassVar, Optional

import numpy as np
import pgvector.sqlalchemy  # type: ignore[import-untyped]
import PIL.Image
import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import catalog, exprs, func
from pixeltable.env import Env

from .base import IndexBase


class EmbeddingIndex(IndexBase):
    """
    Interface to the pgvector access method in Postgres.
    - pgvector converts the cosine metric to '1 - metric' and the inner product metric to '-metric', in order to
      satisfy the Postgres requirement that an index scan requires an ORDER BY ... ASC clause
    - similarity_clause() converts those metrics back to their original form; it is used in expressions outside
      the Order By clause
    - order_by_clause() is used exclusively in the ORDER BY clause
    - embedding function parameters are named '<type-name>_embed', where type-name is ColumnType.Type.name
    """

    class Metric(enum.Enum):
        COSINE = 1
        IP = 2
        L2 = 3

    PGVECTOR_OPS: ClassVar[dict[Metric, str]] = {
        Metric.COSINE: 'vector_cosine_ops',
        Metric.IP: 'vector_ip_ops',
        Metric.L2: 'vector_l2_ops',
    }

    metric: Metric
    value_expr: exprs.FunctionCall
    string_embed: Optional[func.Function]
    image_embed: Optional[func.Function]
    string_embed_signature_idx: int
    image_embed_signature_idx: int
    index_col_type: pgvector.sqlalchemy.Vector

    def __init__(
        self,
        c: catalog.Column,
        metric: str,
        embed: Optional[func.Function] = None,
        string_embed: Optional[func.Function] = None,
        image_embed: Optional[func.Function] = None,
    ):
        if embed is None and string_embed is None and image_embed is None:
            raise excs.Error('At least one of `embed`, `string_embed`, or `image_embed` must be specified')
        metric_names = [m.name.lower() for m in self.Metric]
        if metric.lower() not in metric_names:
            raise excs.Error(f'Invalid metric {metric}, must be one of {metric_names}')
        if not c.col_type.is_string_type() and not c.col_type.is_image_type():
            raise excs.Error('Embedding index requires string or image column')

        self.string_embed = None
        self.image_embed = None

        # Resolve the specific embedding functions corresponding to the user-provided `string_embed`, `image_embed`,
        # and/or `embed`. For string embeddings, `string_embed` will be used if specified; otherwise, `embed` will
        # be used as a fallback, if it has a matching signature. Likewise for image embeddings.

        if string_embed is not None:
            # `string_embed` is specified; it MUST be valid.
            self.string_embed = self._resolve_embedding_fn(string_embed, ts.ColumnType.Type.STRING)
            if self.string_embed is None:
                raise excs.Error(
                    f'The function `{string_embed.name}` is not a valid string embedding: '
                    'it must take a single string parameter'
                )
        elif embed is not None:
            # `embed` is specified; see if it has a string signature.
            self.string_embed = self._resolve_embedding_fn(embed, ts.ColumnType.Type.STRING)

        if image_embed is not None:
            # `image_embed` is specified; it MUST be valid.
            self.image_embed = self._resolve_embedding_fn(image_embed, ts.ColumnType.Type.IMAGE)
            if self.image_embed is None:
                raise excs.Error(
                    f'The function `{image_embed.name}` is not a valid image embedding: '
                    'it must take a single image parameter'
                )
        elif embed is not None:
            # `embed` is specified; see if it has an image signature.
            self.image_embed = self._resolve_embedding_fn(embed, ts.ColumnType.Type.IMAGE)

        if self.string_embed is None and self.image_embed is None:
            # No string OR image signature was found. This can only happen if `embed` was specified and
            # contains no matching signatures.
            assert embed is not None
            raise excs.Error(
                f'The function `{embed.name}` is not a valid embedding: it must take a single string or image parameter'
            )

        # Now validate the return types of the embedding functions.

        if self.string_embed is not None:
            self._validate_embedding_fn(self.string_embed)

        if self.image_embed is not None:
            self._validate_embedding_fn(self.image_embed)

        if c.col_type.is_string_type() and self.string_embed is None:
            raise excs.Error(f"Text embedding function is required for column {c.name} (parameter 'string_embed')")
        if c.col_type.is_image_type() and self.image_embed is None:
            raise excs.Error(f"Image embedding function is required for column {c.name} (parameter 'image_embed')")

        self.metric = self.Metric[metric.upper()]
        self.value_expr = (
            self.string_embed(exprs.ColumnRef(c))
            if c.col_type.is_string_type()
            else self.image_embed(exprs.ColumnRef(c))
        )
        assert isinstance(self.value_expr.col_type, ts.ArrayType)
        vector_size = self.value_expr.col_type.shape[0]
        assert vector_size is not None
        self.index_col_type = pgvector.sqlalchemy.Vector(vector_size)

    def index_value_expr(self) -> exprs.Expr:
        """Return expression that computes the value that goes into the index"""
        return self.value_expr

    def records_value_errors(self) -> bool:
        return True

    def index_sa_type(self) -> sql.types.TypeEngine:
        """Return the sqlalchemy type of the index value column"""
        return self.index_col_type

    def create_index(self, index_name: str, index_value_col: catalog.Column) -> None:
        """Create the index on the index value column"""
        idx = sql.Index(
            index_name,
            index_value_col.sa_col,
            postgresql_using='hnsw',
            postgresql_with={'m': 16, 'ef_construction': 64},
            postgresql_ops={index_value_col.sa_col.name: self.PGVECTOR_OPS[self.metric]},
        )
        conn = Env.get().conn
        idx.create(bind=conn)

    def similarity_clause(self, val_column: catalog.Column, item: Any) -> sql.ColumnElement:
        """Create a ColumnElement that represents '<val_column> <op> <item>'"""
        assert isinstance(item, (str, PIL.Image.Image))
        if isinstance(item, str):
            assert self.string_embed is not None
            embedding = self.string_embed.exec([item], {})
        if isinstance(item, PIL.Image.Image):
            assert self.image_embed is not None
            embedding = self.image_embed.exec([item], {})

        if self.metric == self.Metric.COSINE:
            return val_column.sa_col.cosine_distance(embedding) * -1 + 1
        elif self.metric == self.Metric.IP:
            return val_column.sa_col.max_inner_product(embedding) * -1
        else:
            assert self.metric == self.Metric.L2
            return val_column.sa_col.l2_distance(embedding)

    def order_by_clause(self, val_column: catalog.Column, item: Any, is_asc: bool) -> sql.ColumnElement:
        """Create a ColumnElement that is used in an ORDER BY clause"""
        assert isinstance(item, (str, PIL.Image.Image))
        embedding: Optional[np.ndarray] = None
        if isinstance(item, str):
            assert self.string_embed is not None
            embedding = self.string_embed.exec([item], {})
        if isinstance(item, PIL.Image.Image):
            assert self.image_embed is not None
            embedding = self.image_embed.exec([item], {})
        assert embedding is not None

        if self.metric == self.Metric.COSINE:
            result = val_column.sa_col.cosine_distance(embedding)
            result = result.desc() if is_asc else result
        elif self.metric == self.Metric.IP:
            result = val_column.sa_col.max_inner_product(embedding)
            result = result.desc() if is_asc else result
        else:
            assert self.metric == self.Metric.L2
            result = val_column.sa_col.l2_distance(embedding)
        return result

    @classmethod
    def display_name(cls) -> str:
        return 'embedding'

    @classmethod
    def _resolve_embedding_fn(
        cls, embed_fn: func.Function, expected_type: ts.ColumnType.Type
    ) -> Optional[func.Function]:
        """Find an overload resolution for `embed_fn` that matches the given type."""
        assert isinstance(embed_fn, func.Function)
        for resolved_fn in embed_fn._resolved_fns:
            # The embedding function must be a 1-ary function of the correct type. But it's ok if the function signature
            # has more than one parameter, as long as it has at most one *required* parameter.
            sig = resolved_fn.signature
            if (
                len(sig.parameters) >= 1
                and len(sig.required_parameters) <= 1
                and sig.parameters_by_pos[0].col_type.type_enum == expected_type
            ):
                # We found a valid signature. Now, if it has more than one parameter, we need to transform it into a
                # 1-ary function by fixing all the other parameters to their defaults. This is to ensure that
                # conditional_return_type resolves correctly.
                if len(sig.parameters) == 1:
                    unary_fn = resolved_fn
                else:
                    assert all(sig.parameters_by_pos[i].has_default for i in range(1, len(sig.parameters)))
                    defaults = {param.name: param.default for param in sig.parameters_by_pos[1:]}
                    unary_fn = resolved_fn.using(**defaults)
                assert not unary_fn.is_polymorphic
                assert len(unary_fn.signature.parameters) == 1
                return unary_fn
        return None

    @classmethod
    def _validate_embedding_fn(cls, embed_fn: func.Function) -> None:
        """Validate the given embedding function."""
        assert not embed_fn.is_polymorphic

        return_type = embed_fn.signature.return_type

        if not isinstance(return_type, ts.ArrayType):
            raise excs.Error(
                f'The function `{embed_fn.name}` is not a valid embedding: '
                f'it must return an array, but returns {return_type}'
            )

        shape = return_type.shape
        if len(shape) != 1 or shape[0] is None:
            raise excs.Error(
                f'The function `{embed_fn.name}` is not a valid embedding: '
                f'it must return a 1-dimensional array of a specific length, but returns {return_type}'
            )

    def as_dict(self) -> dict:
        return {
            'metric': self.metric.name.lower(),
            'string_embed': None if self.string_embed is None else self.string_embed.as_dict(),
            'image_embed': None if self.image_embed is None else self.image_embed.as_dict(),
        }

    @classmethod
    def from_dict(cls, c: catalog.Column, d: dict) -> EmbeddingIndex:
        string_embed = func.Function.from_dict(d['string_embed']) if d['string_embed'] is not None else None
        image_embed = func.Function.from_dict(d['image_embed']) if d['image_embed'] is not None else None
        return cls(c, metric=d['metric'], string_embed=string_embed, image_embed=image_embed)



from typing import Any


def get_client(**kwargs: Any) -> Any:
    import boto3
    import botocore

    try:
        boto3.Session().get_credentials().get_frozen_credentials()
        config = botocore.config.Config(**kwargs)
        return boto3.client('s3', config=config)  # credentials are available
    except AttributeError:
        # No credentials available, use unsigned mode
        config_args = kwargs.copy()
        config_args['signature_version'] = botocore.UNSIGNED
        config = botocore.config.Config(**config_args)
        return boto3.client('s3', config=config)
